;;; emacs-process.el --- list processes by Emacs

;; Copyright (C) 2010, 2011  Leo

;; Author: Leo <sdl.web@gmail.com>
;; Keywords: processes, tools

;; This file is not part of GNU Emacs.

;; GNU Emacs is free software: you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation, either version 3 of the License, or
;; (at your option) any later version.

;; GNU Emacs is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GNU Emacs.  If not, see <http://www.gnu.org/licenses/>.

;;; Commentary:

;;; My attempt to move the primitive list-processes to elisp.

;;; Code:

(eval-when-compile (require 'cl))

(defun emacs-process-info (&optional query-only)
  "Return a list of plist of process information."
  (delete nil
          (mapcar (lambda (p)
                    (when (or (not query-only)
                              (and query-only (process-query-on-exit-flag p)))
                      (list :proc (process-name p)
                            :status (symbol-name (process-status p))
                            :buffer (let ((buf (process-buffer p)))
                                      (cond
                                       ((null buf) "(none)")
                                       ((not (buffer-live-p buf)) "(Killed)")
                                       (t (buffer-name buf))))
                            :tty (process-tty-name p)
                            :command (case (process-type p)
                                       (network (destructuring-bind
                                                    (&key server type service host
                                                          &allow-other-keys)
                                                    (process-contact p t)
                                                  (format "(network %s %s)"
                                                          (if type "datagram" "stream")
                                                          (if server
                                                              (format "server on %s" service)
                                                            (format "connection to %s" host)))))
                                       (serial (destructuring-bind
                                                   (&key port speed &allow-other-keys)
                                                   (process-contact p t)
                                                 (format "(serial port %s%s)"
                                                         (or port "?")
                                                         (if speed
                                                             (format " at %s b/s" speed)
                                                           ""))))
                                       (otherwise (mapconcat 'identity (process-command p) " "))))))
                  (process-list))))

(defun emacs-process-compute-widths (info)
  "Compute the width of each column."
  (loop for (key value) on (car info) by 'cddr
        for pwidth = (loop for i in info
                           maximize (length (plist-get i key)))
        collect (if (zerop pwidth)
                    0
                  (max pwidth (1- (length (symbol-name key)))))))

(defun emacs-process-insert (info)
  (let ((widths (emacs-process-compute-widths info))
        (index 0) beg)
    (while info
      (incf index)
      (setq beg (point))
      (loop for (k v) on (car info) by 'cddr
            ;; the key of the plist is used as heading
            for h = (substring (symbol-name k) 1)
            ;; use a small nonzero value for last column
            for w in (append (butlast widths) '(1))
            for curcol = (current-column)
            ;; don't display column with width zero
            unless (zerop w)
            do
            (case index
              (1 (insert (capitalize h)))
              (2 (insert (make-string (length h) ?-)))
              (otherwise (and v (insert v))))
            (indent-to (+ curcol w))
            ;; insert a space between each column
            (insert " "))
      (insert "\n")
      (when (> index 2)
        (pop info)))))

;;;###autoload
(when (subrp (symbol-function 'list-processes))
  (fset 'list-processes-primitive (symbol-function 'list-processes))
  (defalias 'list-processes 'emacs-process-list))

;;;###autoload
(defun emacs-process-list (&optional query-only)
  "Display a list of all processes.
If optional argument QUERY-ONLY is non-nil, only processes with
the query-on-exit flag set will be listed.
Any process listed as exited, signaled or closed is actually 
eliminated after the listing is made."
  (interactive "P")
  (let ((info (emacs-process-info query-only))
        (inhibit-read-only t)
        (bufname "*Process List*")
        line)
    (if (not info)
        (message "Process list empty")
      (with-current-buffer (get-buffer-create bufname)
          (help-mode)
          (setq line (line-number-at-pos))
          (erase-buffer)
          (setq truncate-lines t
                buffer-undo-list t)
          (emacs-process-insert info)
          (set-buffer-modified-p nil)
          (goto-char (point-min))
          (forward-line (1- line))
          (display-buffer (current-buffer)))
      (mapc (lambda (p)
              (when (memq (process-status p) '(exit signal closed))
                (delete-process p)))
            (process-list)))))

(provide 'emacs-process)
